#!/usr/bin/env python

from functools import wraps
from itertools import count
from time import sleep
from typing import (
    Any, Callable, Generator, Iterable, Optional, Tuple, Type, Union)

from . import func
from . import fs


def uniq(a: Iterable) -> Generator:
    '''
    Generate the argument iterator with the duplicate members removed.

    >>> list(uniq([1, 2, 2, 3]))
    [1, 2, 3]
    '''
    seen: set = set()
    for e in a:
        if e not in seen:
            seen.add(e)
            yield e
        else:
            continue


def retry(
        times: Optional[int]=None,
        wait: int=0,
        exceptions: Union[Type[Exception], Tuple[Type[Exception]]]=Exception,
        default: Any=None) -> Callable:
    '''
    Returns a version of `f' that will automatically retry `times' times if the
    invokation raises any of the given exceptions. Will wait `wait' seconds to
    try again. If `times' is None, will retry indefinitely. Otherwise, when
    retry has reached that many of attempts, it will return `default' or raise
    ValueError if default is None.

    To be used as a decorator, parameters should be given as keyword arguments.

    >>> import requests as r
    >>> fetch = retry(r.get, times=3, exceptions=ConnectionError)
    >>> fetch(URL)
    <Response [200]>
    '''
    def _impl(f):
        @wraps(f)
        def _wrapper(*args, **kwargs):
            attempts = count() if times is None else range(times)
            first_pass = True
            for _ in attempts:
                if not first_pass:
                    sleep(wait)
                first_pass = False
                try:
                    return f(*args, **kwargs)
                except exceptions:
                    pass
            if default is not None:
                return default
            else:
                raise ValueError('Function application failed')
        return _wrapper
    return _impl



#!/usr/bin/env python

'''
wbutil/queue.py

Helpers which wrap standard library threaded queue.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

from queue import Queue as StdQueue
from threading import Thread


class Queue(StdQueue):
    '''
    Simplify the interface provided by standard library queue. Specifically
    tailored to processing lists of data.
    '''

    def __init__(self, func, maxsize=0, nthreads=4):
        super().__init__(maxsize)
        self._func = func
        self._results = {}
        self._threads = [None] * nthreads
        self._nthreads = nthreads

    def _worker(self):
        while True:
            key, item = self.get()
            if item is None:
                break
            self._results[key] = self._func(item)
            self.task_done()

    def __enter__(self):
        '''Start threads.'''
        for i in range(self._nthreads):
            self._threads[i] = Thread(target=self._worker)
            self._threads[i].start()
        return self

    def __exit__(self, exception_t, exception, traceback):
        '''Stop and join the threads.'''
        for i in range(self._nthreads):
            self.put((0, None))
        for t in self._threads:
            t.join()

    def map(self, data_sequence):
        '''A special case where the result becomes a list.'''
        self._results = [None] * len(data_sequence)
        for item in enumerate(data_sequence):
            self.put(item)
        self.join()
        return self._results

    # Nice aliases for context functions, for use when `with' not appropriate
    start = __enter__
    shutdown = __exit__



#!/usr/bin/env python

'''
wbutil/func.py

Functional-programming-oriented utilities.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

from functools import partial, reduce
from inspect import signature
from itertools import chain
from typing import Any, Callable, Iterable

__all__ = [
    'compose',
    'partialright',
    'starcompose',
    'smartcompose',
]

_UnaryFunc = Callable[[Any], Any]
_NAryFunc = Callable[[Iterable[Any]], Iterable[Any]]


class compose(object):
    '''
    Returns a function which calls the argument functions in order from left
    to right (first to last). Undefined behavior for non-unary functions.

    >>> from functools import partial; from operator import mul
    >>> times2tostr = compse(partial(mul, 2), str)
    >>> times2tostr(10)
    '20'
    '''

    def __init__(self, *funcs: _UnaryFunc) -> None:
        self.funcs = funcs

    def __call__(self, arg: Any) -> Any:
        '''Invoke the composed pipeline with the specified argument.'''
        return reduce(lambda acc, f: f(acc), self.funcs, arg)

    def __reversed__(self) -> 'compose':
        '''
        Gives a composition with the calling order reversed. Allows compose to
        emulate traditional compositional ordering.
        '''
        funcs = tuple(reversed(self.funcs))
        return type(self)(*funcs)


class starcompose(compose):
    '''
    Returns a function of *args that applies the argument functions in
    order from left to right (first to last). All supplied frunctions must =
    return an iterable and accept *args.

    >>> switch = lambda a, b: (b, a)
    >>> tup2str = lambda a, b: (str(a), str(b))
    >>> switchandstr = starcompose(switch, tup2str)
    >>> switchandstr(1, 2)
    ('2', '1')
    '''

    # Overridden to annotate n-ary function support
    def __init__(self, *funcs: _NAryFunc) -> None:
        super().__init__(*funcs)

    def __call__(self, *args: Any) -> Any:
        '''Invoke the composed pipeline with unpacking.'''
        return reduce(lambda acc, f: f(*acc), self.funcs, args)


class smartcompose(compose):
    '''
    WIP

    Represents the composition of a list of functions, appied in order from
    left to right. Uses inspection to determine whether to use a regular or
    unpacked application.

    >>>
    TODO: complete example
    '''

    # Overridden to annotate support for any function
    def __init__(self, *funcs: Callable) -> None:
        super().__init__(*funcs)

    def __call__(self, *args: Any) -> Any:
        '''
        Invoke the pipeline, deciding at each step whether to unpack arguments.
        '''
        acc = args
        for f in self.funcs:
            try:
                sig = signature(f)
            except ValueError:
                continue
            if len(sig.parameters) == 0:
                acc = f()
            elif len(sig.parameters) == 1:
                acc = f(acc)
            else:
                acc = f(*acc)
        return acc


class partialright(partial):
    '''
    Returns a new function with arguments partially applied from right to left.

    >>> data = [1, 2, 3, 4]
    >>> process_data = partialright(map, data)
    >>> stringified = process_data(str)
    >>> stringified
    ['1', '2', '3', '4']
    '''

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        '''Call self as a function.'''
        pos_args = chain(args, reversed(self.args))
        keywords = self.keywords.copy()
        keywords.update(kwargs)
        return self.func(*pos_args, **keywords)



#!/usr/bin/env python

'''
wbutil/fs.py

Filesystem utilities.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

from typing import Any, Callable, TextIO, TypeVar

__all__ = [
    'saveobj',
    'tryopen',
]


def saveobj(constructor: Callable[[], Any], path: str) -> Any:
    '''
    Construct an object and save it to the disk for later use. Useful for
    objects with expensive constructors such as ones that include Web requests
    since they're only run if the storage file doesn't already exist.

    >>> from functools import partial; from requests import get
    >>> req = objstorage(partial(get, 'http://example.com'), 'myreq.pickle')
    >>> req
    <Response [200]>
    '''
    from os.path import exists
    from pickle import dump, load
    if not exists(path):
        obj = constructor()
        with open(path, 'wb') as fs:
            dump(obj, fs)
    else:
        with open(path, 'rb') as fs:
            obj = load(fs)
    return obj


_ProcessReturn_t = TypeVar('_ProcessReturn_t')


def tryopen(
        path: str,
        process: Callable[[TextIO], _ProcessReturn_t]=lambda fs: fs.read(),
        default: Any=None) -> _ProcessReturn_t:
    '''
    Attempt to open the file at `path`, process it, and close it. If file
    cannot be loaded, the default value is returned. Does not swallow errors
    from the processing function.

    >>> from json import load
    >>> data = try_open('data.json', process=load, default={})
    >>> data
    {'my': 'data'}
    >>> description = try_open('doesnt_exist.txt', default='No description')
    >>> description
    'No description'
    '''
    try:
        with open(path, 'r') as fs:
            return process(fs)
    except OSError:
        return default



#!/usr/bin/env python

'''
setup.py

wbutil-py package metadata.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

from setuptools import setup, find_packages


setup(name='wbutil',
      version='0.0.1a5',
      packages=find_packages(),
      url='https://github.com/wbadart/wbutil-py',

      author='Will Badart',
      author_email='wbadart@live.com',
      description='Python utility library',
      license='MIT',
      keywords='utility util helper library',
      homepage='https://github.com/wbadart/wbutil-py',
      project_urls={
          'Bug Tracker': 'https://github.com/wbadart/wbutil-py/issues',
          'Documentation': 'https://github.com/wbadart/wbutil-py/wiki',
          'Source Code': 'https://github.com/wbadart/wbutil-py',
      },

      zip_safe=False)
